'''
Definition of the half-Rabi frequency and all other functions needed to calculate it
'''

import numpy as np
from state import *
from constants import *
from sympy.physics.wigner import wigner_3j
from sympy.physics.wigner import wigner_6j

# Intensity of Gaussian TEM_00 laser beam profile
# Inputs: P_las is power of laser given by power meter in mW, r_sigma is radius at 2D std in mm, r is radius in mm
# Output: intensity at radius r from beam axis given a Gaussian intesnity profile
def gaussianIntensity(P_las, r_sigma, r):
    return (P_las/(2*PI*r_sigma*r_sigma))*np.exp(-(r*r)/(2*r_sigma*r_sigma))

# Polarisation quantum number
def quantumPolarisation(q):
    if q == 1:
        return 0
    if (q == -1) or (q == 0):
        return 1

# Coupling constants
# Input: e and g are State objects, q is the laser polarisation
# Output: the coupling coefficient which is the Dipole matrix element
def coupling(e, g, q):
    sign = np.power(-1, quantumPolarisation(q)+ e.S + 1 + e.L + g.J + e.J - e.m)
    factor = np.sqrt((2*e.J+1)*(2*e.L+1)*(2*g.J+1))
    wig6j = wigner_6j(e.L, e.J, e.S, g.J, g.L, 1)
    wig3j = wigner_3j(e.J, 1, g.J, -1*e.m, q, g.m)
    return sign*factor*wig6j*wig3j

# Caluclates the half-Rabi frequency in GHz.
# Inputs: intensity given in mW/mm^2, lifetime in ns/rad, wavelength of laser in m
# Output: the half-Rabi frequency constant (without coupling coefficient) in GHz
def halfRabiFreq(intensity, lifetime, wavelength):
    I = intensity*1000  # Convert mW/mm^2 to W/m^2
    h = 6.6260715e-34  # Plank's constant
    c = 299792458  # Speed of light
    tau = lifetime*1e-9  # Convert ns/rad lfetime to s/rad
    
    return np.sqrt((3*I*wavelength**3)/(8*np.pi*c*h*tau))*1e-9  #Gives half-Rabi freq in Grad/s


'''
Define functions for the detuning of an atomic system
'''
from state import *

# Detunings
# Input: e is a State object, g is a State object
def detuning(e, g):
    return e.w - g.w
    
# Doppler detuning.
# Input: w_q is the angular frequency of the laser in rad/s, lambda_q is the wavelength of light in m
# ,v_z is the velocity component of atoms in direction of laser, e and g are State objects
def dopplerDelta(e, g, w_q, lambda_q, v_z):
    return w_q - v_z/lambda_q + e.w - g.w


'''
This is a file to define functions for rotating density matrices in the QED simulation of
a laser-atom system.
Author: Manish Patel
Date created: 12/05/2021
'''

from density_matrix import *
import numpy as np
import math
import copy

def wigner_D(J, alpha, beta, gamma):
    '''
    Calculates the Wigner D-matrix for rotation by Eueler angles (alpha, beta, gamma).
    Inputs:
        J: total angular momentum quantum number of the state which will be rotated with the 
        resulting D-matrix
        alpha: rotation around z-axis
        beta: rotation about the y'-axis
        gamma: rotation about the z''-axis
    Returns:
        A square matrix of size 2J+1
    '''
    size = 2*J+1  # Number of sub-states
    m = np.linspace(-J, J, size, dtype=int)  # Projections of J
    D = np.zeros((size, size), dtype = np.complex)  # Set up D-matrix
    for i, mp in enumerate(m):
        for j, mpp in enumerate(m):
            alpha_const = math.cos(-mp*alpha)+1.j*math.sin(-mp*alpha)
            gamma_const = math.cos(-mpp*gamma)+1.j*math.sin(-mpp*gamma)
            D[i, j] = alpha_const*small_Wigner_D(J, beta, mp, mpp)*gamma_const
    return D

def small_Wigner_D(J, beta, mp, m):
    '''
    Calculates the small Wigner D-matrix elements for rotation by Euler angles (alpha, beta, gamma)
    '''
    const = np.sqrt((math.factorial(J+mp))*math.factorial(J-mp)*math.factorial(J+m)*math.factorial(J-m))
    d_sum = 0
    # Define limits so sum does not contain negative factorials
    s_max = min(J+m, J-mp)
    s_min = max(0, m-mp)
    sum_index = np.linspace(s_min, s_max, )
    for s in range(s_min, s_max+1):  # Have to go to s_max+1 or will miss out on the s_max value
        numerator = np.power(-1, mp - m + s)*np.power(math.cos(beta/2), 2*J+m-mp-2*s)*np.power(math.sin(beta/2), mp-m+2*s)
        denominator = math.factorial(J+m-s)*math.factorial(s)*math.factorial(mp-m+s)*math.factorial(J-mp-s)
        d_sum += numerator/denominator
    
    return const*d_sum

'''
Rotates a density matrix by the wigner D-matrix determined by Euler angles (alpha, beta, gamma).
(rho_newframe) = (D)(rho)(D*). The z-y-z convention is used where alpha rotates around the z-axis,
then beta rotates around the new y-axis, and gamma rotates around the new z-axis
'''
def rotation(rho, J, alpha, beta, gamma):
    D_matrix = np.transpose(wigner_D(J, alpha, beta, gamma))
    D_conj = np.transpose(np.conj(D_matrix))
    return np.dot(D_matrix, np.dot(rho, D_conj))


'''
Rotate the excited and ground state populations by the Euler angles alpha, beta, gamma
'''
def rotateInitialMatrix(flat_rho, n, E, G, alpha, beta, gamma):
    # Make a copy to return
    rotated_rho = copy.deepcopy(flat_rho)
    # Rotate the excited state populations and atomic coherences
    J_E = JNumber(E)
    rotated_excited_rho = rotation(getSingleStateMatrix(rotated_rho, n, E), J_E, alpha, beta, gamma)
    appendDensityMatrixToFlatCoupledMatrix(rotated_rho, rotated_excited_rho, E, n)
    
    # Rotate the ground state populations and atomic coherences
    J_G = JNumber(G)
    rotated_ground_rho = rotation(getSingleStateMatrix(rotated_rho, n, G), J_G, alpha, beta, gamma)
    appendDensityMatrixToFlatCoupledMatrix(rotated_rho, rotated_ground_rho, G, n)
    
    return rotated_rho


'''
Density matrix operations
Author: Manish Patel
Date Created: 11/06/2021
'''

import numpy as np
from index import *

def getSingleStateMatrix(flat_rho, n, sub_states):
    """Obtain an angular momentum state density matrix from the flattened coupled state density rho vector
    Inputs:
    flat_rho : array of arrays with one column of all density matrix elements of coupled E & G states
    n : number of states in total laser-coupled system
    sub_states : a list of the excited or ground states, E or G respectively
    Returns:
    A square matrix of size length of sub_states. Elements are ordered from left to right
    according to the order of sub_states e.g. if state labelled 1 is first then first element
    would correspond to rho_11 i.e. population of state 1
    """
    # Set up the state matrix
    state_density_matrix = np.zeros((len(sub_states), len(sub_states)), dtype = np.complex)
    
    # Populate matrix
    for i, sub_state in enumerate(sub_states):
        for j, sub_state_p in enumerate(sub_states):
            state_density_matrix[i, j] = flat_rho[index(sub_state, sub_state_p, n), 0]
    return state_density_matrix


def JNumber(state_list):
    """
    Calculate the angular momentum from the number of states in the list
    """
    return int((len(state_list)-1)/2)  # J = (m-1)/2


def appendDensityMatrixToFlatCoupledMatrix(flatrho, density_rho, sub_states, n):
    """
    Adds density matrix elements to a flat, coupled matrix.
    Inputs:
    - flat_rho: array of arrays with one column of all density matrix elements of coupled E and G states
    - n: number of states in total laser-coupled system
    - sub-states: a list of the excited or ground states
    - density_rho: either the excited or ground state density matrix with the convention that the 
                   upper left-hand of the matrix is state population for m_J = -J if the state has angular momentum J 
"""
    for i, sub_state in enumerate(sub_states):
        for j, sub_state_p in enumerate(sub_states):
            flatrho[index(sub_state, sub_state_p, n), 0] = density_rho[i, j]         



'''
This file contains the function to calculate the evolution of the density matrix
for an atomic system interacting with a laser
'''

from constants import *
import numpy as np
from detuning import *
from half_rabi_freq import *
from matrix_methods import *
from time_evolution_matrix import *
import scipy.sparse.linalg as sparsela

def timeEvolution(n, E, G, Q, Q_decay, tau, laser_intensity, laser_wavelength, time, rho0, rho_output, tau_f = None, rabi_scaling = None, print_eq = None, pretty_print_eq = None, atomic_velocity = None):
    
    rabi = halfRabiFreq(laser_intensity, tau, laser_wavelength)
            
    A = timeEvolutionMatrix(n, E, G, Q, Q_decay, tau, laser_wavelength, laser_intensity, 
                        tau_f = tau_f, rabi_scaling = rabi_scaling, symbolic_print = pretty_print_eq, 
                                    numeric_print = print_eq, atomic_velocity = atomic_velocity)

    # Compute the diagonalised matrix D and matrix of eigenvectors V
    D = diagonalise(A)
    V = matrixOfEigenvec(A)
    f = np.dot(la.inv(V), rho0)  # Compute V^-1*rho(0)
    
    # Calculate the exponential
    for position, t in enumerate(time, start = 0):
        # Use expm() which computes the matrix exponential using the Pade approximation
        VexpDt = np.dot(V, sparsela.expm(D*t))  # Use sparse linear algebra to compute e^Dt as this is faster
        rho_t = np.dot(VexpDt, f)
        
        # Append density matrix elements
        # rho(t)_ee
        for e in E:
            for ep in E:
                rho_output[index(e, ep, n)][position] = rho_t[index(e, ep, n), 0]
        # rho(t)_gg
        for g in G:
            for gp in G:
                rho_output[index(g, gp, n)][position] = rho_t[index(g, gp, n), 0]
        # rho(t)_eg
        for e in E:
            for g in G:
                rho_output[index(e, g, n)][position] = rho_t[index(e, g, n), 0]
        #rho(t)_ge
        for g in G:
            for e in E:
                rho_output[index(g, e, n)][position] = rho_t[index(g, e, n), 0]

def timeEvolutionDopplerAveraging(n, E, G, Q, Q_decay, tau, laser_intensity, laser_wavelength, doppler_width, doppler_detunings, time, rho0, rho_output, tau_f = None, rabi_scaling = None, print_eq = None, pretty_print_eq = None, atomic_velocity = None):
    
    if(print_eq or pretty_print_eq != None):
        print("Cannot print equations when beam profile or doppler averaging!")
        print_eq = None
        pretty_print_eq = None
    
    # Calculate doppler_spacing
    d_doppler = abs(doppler_detunings[1] - doppler_detunings[0])
    doppler_factor = d_doppler/(np.sqrt(2*PI*doppler_width*doppler_width))
    
    for doppler_delta in doppler_detunings:
            # Calculate the half-rabi frequency
            rabi = halfRabiFreq(laser_intensity, tau, laser_wavelength)
            
            A = timeEvolutionMatrix(n, E, G, Q, Q_decay, tau, laser_wavelength, laser_intensity, 
                        tau_f = tau_f, rabi_scaling = rabi_scaling, symbolic_print = pretty_print_eq, 
                                    numeric_print = print_eq, atomic_velocity = atomic_velocity)

            # Compute the diagonalised matrix D and matrix of eigenvectors V
            D = diagonalise(A)
            V = matrixOfEigenvec(A)
            f = np.dot(la.inv(V), rho0)  # Compute V^-1*rho(0)

            # Calculate the exponential
            for position, t in enumerate(time, start = 0):
                # Use expm() which computes the matrix exponential using the Pade approximation
                VexpDt = np.dot(V, sparsela.expm(D*t))  # Use sparse linear algebra to compute e^Dt as this is faster
                rho_t = np.dot(VexpDt, f)

                # Append density matrix for each ring and each detuning fraction
                # rho(t)_ee
                for e in E:
                    for ep in E:
                        rho_output[index(e, ep, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(rho_t[index(e, ep, n), 0])
                # rho(t)_gg
                for g in G:
                    for gp in G:
                        rho_output[index(g, gp, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(rho_t[index(g, gp, n), 0])
                # rho(t)_eg
                for e in E:
                    for g in G:
                        rho_output[index(e, g, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(rho_t[index(e, g, n), 0])
                #rho(t)_ge
                for g in G:
                    for e in E:
                        rho_output[index(g, e, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(rho_t[index(e, g, n), 0])
                        
def timeEvolutionGaussianAveraging(n, E, G, Q, Q_decay, tau, laser_power, r_sigma, n_intensity, laser_wavelength, time, rho0, rho_output, tau_f = None, rabi_scaling = None, print_eq = None, pretty_print_eq = None, atomic_velocity = None):
    
    if(print_eq or pretty_print_eq != None):
        print("Cannot print equations when beam profile or doppler averaging!")
        print_eq = None
        pretty_print_eq = None
    
    # Create rings of laser beam to integrate populations over
    R = np.linspace(r_sigma, 3*r_sigma, n_intensity)
    
    for k, r in enumerate(R, start = 0):
            # Calculate the half-rabi frequency
            laser_intensity = gaussianIntensity(laser_power, r_sigma, r)
            rabi = halfRabiFreq(laser_intensity, tau, laser_wavelength)
            
            A = timeEvolutionMatrix(n, E, G, Q, Q_decay, tau, laser_wavelength, laser_intensity, 
                        tau_f = tau_f, rabi_scaling = rabi_scaling, symbolic_print = pretty_print_eq, 
                                    numeric_print = print_eq, atomic_velocity = atomic_velocity)

            # Compute the diagonalised matrix D and matrix of eigenvectors V
            D = diagonalise(A)
            V = matrixOfEigenvec(A)
            f = np.dot(la.inv(V), rho0)  # Compute V^-1*rho(0)

            # Calculate the exponential
            for position, t in enumerate(time, start = 0):
                # Use expm() which computes the matrix exponential using the Pade approximation
                VexpDt = np.dot(V, sparsela.expm(D*t))  # Use sparse linear algebra to compute e^Dt as this is faster
                rho_t = np.dot(VexpDt, f)

                # Append density matrix for each ring
                # rho(t)_ee
                for e in E:
                    for ep in E:
                        rho_output[index(e, ep, n)][position] += (2*k+1)*(rho_t[index(e, ep, n), 0])/(n_intensity*n_intensity)
                # rho(t)_gg
                for g in G:
                    for gp in G:
                        rho_output[index(g, gp, n)][position] += (2*k+1)*(rho_t[index(g, gp, n), 0])/(n_intensity*n_intensity)
                # rho(t)_eg
                for e in E:
                    for g in G:
                        rho_output[index(e, g, n)][position] += (2*k+1)*(rho_t[index(e, g, n), 0])/(n_intensity*n_intensity)
                #rho(t)_ge
                for g in G:
                    for e in E:
                        rho_output[index(g, e, n)][position] += (2*k+1)*(rho_t[index(e, g, n), 0])/(n_intensity*n_intensity)
    

def timeEvolutionGaussianAndDopplerAveraging(n, E, G, Q, Q_decay, tau, laser_power, r_sigma, n_intensity, laser_wavelength, doppler_width, doppler_detunings, time, rho0, rho_output, tau_f = None, rabi_scaling = None, print_eq = None, pretty_print_eq = None, atomic_velocity = None):
    
    if(print_eq or pretty_print_eq != None):
        print("Cannot print equations when beam profile or doppler averaging!")
        print_eq = None
        pretty_print_eq = None
    
    # Create rings of laser beam to integrate populations over
    R = np.linspace(r_sigma, 3*r_sigma, n_intensity)
    
    # Calculate doppler_spacing
    d_doppler = abs(doppler_detunings[1] - doppler_detunings[0])
    doppler_factor = d_doppler/(np.sqrt(2*PI*doppler_width*doppler_width))
    
    for doppler_delta in doppler_detunings:
        for k, r in enumerate(R, start = 0):
            # Calculate the half-rabi frequency
            laser_intensity = gaussianIntensity(laser_power, r_sigma, r)
            rabi = halfRabiFreq(laser_intensity, tau, laser_wavelength)
            
            A = timeEvolutionMatrix(n, E, G, Q, Q_decay, tau, laser_wavelength, laser_intensity, 
                        tau_f = tau_f, rabi_scaling = rabi_scaling, symbolic_print = pretty_print_eq, 
                                    numeric_print = print_eq, atomic_velocity = atomic_velocity)

            # Compute the diagonalised matrix D and matrix of eigenvectors V
            D = diagonalise(A)
            V = matrixOfEigenvec(A)
            f = np.dot(la.inv(V), rho0)  # Compute V^-1*rho(0)

            # Calculate the exponential
            for position, t in enumerate(time, start = 0):
                # Use expm() which computes the matrix exponential using the Pade approximation
                VexpDt = np.dot(V, sparsela.expm(D*t))  # Use sparse linear algebra to compute e^Dt as this is faster
                rho_t = np.dot(VexpDt, f)

                # Append density matrix for each ring and each detuning fraction
                # rho(t)_ee
                for e in E:
                    for ep in E:
                        rho_output[index(e, ep, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(2*k+1)*(rho_t[index(e, ep, n), 0])/(n_intensity*n_intensity)
                # rho(t)_gg
                for g in G:
                    for gp in G:
                        rho_output[index(g, gp, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(2*k+1)*(rho_t[index(g, gp, n), 0])/(n_intensity*n_intensity)
                # rho(t)_eg
                for e in E:
                    for g in G:
                        rho_output[index(e, g, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(2*k+1)*(rho_t[index(e, g, n), 0])/(n_intensity*n_intensity)
                #rho(t)_ge
                for g in G:
                    for e in E:
                        rho_output[index(g, e, n)][position] += doppler_factor*np.exp(-np.power(doppler_delta/doppler_width, 2)/2)*(2*k+1)*(rho_t[index(e, g, n), 0])/(n_intensity*n_intensity)
                                                                                                     


# Function definitions for methods which operate on matrices

import numpy as np
import scipy.linalg as la

# Diagonalise a matrix.
# Input: A is a square matrix
# Output: diagonalised A
def diagonalise(A):
    eigenvals, eigenvecs = la.eig(A)
    return np.diag(eigenvals)


# Calculate a matrix of eigenvectors.
# Input: A is a square matrix
# Output: matrix of eigenvectors
def matrixOfEigenvec(A):
    eigenvals, eigenvecs = la.eig(A)
    return eigenvecs


'''
The LaserAtomSystem class definition.
'''

import time_evolution as te
import numpy as np
import index as ix
import rotation as ro
import density_matrix as dm

class LaserAtomSystem:
    """A user-defined laser field acting on a user-defined atomic system
    """
    
    # Class variables
    Q_decay = [1, 0 ,-1]
    rho_t = []
    
    def __init__(self, E, G, tau, Q, laser_wavelength, laser_intensity = None, 
                 laser_power = None, tau_f = None):
        self.E = E  # list of excited States
        self.G = G  # list of ground States
        self.tau = tau  # lifteime in ns/rad, N.B NIST database uses A_ki in rad/s
        self.Q = Q  # laser radiation polarisation
        self.laser_wavelength = laser_wavelength  # wavelength of the laser in nm
        self.tau_f = tau_f  # lifetime of decay to other states (can be non-radiative) in ns/rad
        self.laser_intensity = laser_intensity  # in mW/mm^2
        self.laser_power = laser_power  # in mW
        self.rho_0 = np.zeros((self.n*self.n, 1), dtype = complex)  # flattened density matrix
        
    @property
    def n(self):
        """ Total number of sub-states
        """
        return int(len(self.G)+len(self.E))
    
    @property
    def rho_e0(self):
        """ Upper state density matrix for the initial condition
        """
        return dm.getSingleStateMatrix(self.rho_0, self.n, self.E)
    
    @property
    def rho_g0(self):
        """ Lower state density matrix for the initial condition
        """
        return dm.getSingleStateMatrix(self.rho_0, self.n, self.G)
    
    @property
    def rho_et(self):
        """ Upper state density matrix for all of the time evolution
        """
        rho_et = []
        flipped_rho_t = np.transpose(self.rho_t)  # Flip to loop over all rho
        for rho in flipped_rho_t:
            new_rho = np.zeros((self.n*self.n, 1), dtype = complex)  # Placeholder
            for i, element in enumerate(new_rho):
                new_rho[i, 0] = rho[i]
            rho_et.append(dm.getSingleStateMatrix(new_rho, self.n, self.E))
        return rho_et
    
    @property
    def rho_gt(self):
        """ Lower state density matrix for all of the time evolution
        """
        rho_et = []
        flipped_rho_t = np.transpose(self.rho_t)  # Flip to loop over all rho
        for rho in flipped_rho_t:
            new_rho = np.zeros((self.n*self.n, 1), dtype = complex)  # Placeholder
            for i, element in enumerate(new_rho):
                new_rho[i, 0] = rho[i]
            rho_et.append(dm.getSingleStateMatrix(new_rho, self.n, self.G))
        return rho_et
    
    def __repr__(self):
        E_str = [e.label for e in self.E]
        G_str = [g.label for g in self.G]
        return f"LaserAtomSystem({E_str}, {G_str}, {self.tau}, {self.Q}, {self.Q_decay}, {self.laser_wavelength}, {self.tau_f}, {self.laser_intensity}, {self.laser_power})"
        
    
    def Rho_0(self, i, j):
        """ Accessor for an element in rho_0
        Args:
            i (State): First state index
            j (State): Second state index
        Example:
            print(Rho_0(one, two))
        """
        row = ix.index(i, j, self.n)
        return self.rho_0[row, 0]
    
    def setRho_0(self, i, j, value):
        """Sets a value to an element of rho_0
        """
        if(value > 1):
            print("Cannot set an element of a density matrix > 1!")
            return
        else:
            row = ix.index(i, j, self.n)
            self.rho_0[row, 0] = value
    
    def appendDensityMatrixToRho_0(self, density_rho):
        size = len(density_rho)
        if(size == len(G)):
            sub_states = self.G
        elif(size == len(E)):
            sub_states = self.E
        else:
            print("Size of density_rho does not match with excited or ground states")
            return
        dm.appendDensityMatrixToFlatCoupledMatrix(self.rho_0, density_rho, sub_states, self.n)
            
    
    def clearRho_0(self):
        """Makes all values of rho_0 zero
        """
        self.rho_0 = np.zeros((self.n*self.n, 1), dtype = complex)
    
    def Rho_t(self, i, j):
        """ Accessor for an element in rho_t
        Args:
            i (State): First state index
            j (State): Second state index
        Returns:
            Array of an element in laser-atom system for all of the simulation time
        Example:
            print(Rho_t(one, two))
        """
        return self.rho_t[ix.index(i, j, self.n)]
    
    def rotateRho_0(self, alpha, beta, gamma):
        """ Rotate rho_0 by the Euler angles alpha, beta, and gamma.
        """
        self.rho_0 = ro.rotateInitialMatrix(self.rho_0, self.n, self.E, self.G, alpha, beta, gamma)
    
    def rotateRho_t(self, alpha, beta, gamma):
        """ Rotate rho_0 by the Euler angles alpha, beta, and gamma.
        """
        print("Optical coherences are preserved under rotation. To obtain these in a new reference frame, rotate rho_0 and then evolve in the new reference frame with the correct polarisation.")
        rotated_rho_t = []
        # Flip to loop over all rho
        for rho in np.transpose(self.rho_t):
            new_rho = np.zeros((self.n*self.n, 1), dtype = complex)  # Placeholder
            for i, element in enumerate(new_rho):
                new_rho[i, 0] = rho[i]
            new_rho = ro.rotateInitialMatrix(new_rho, self.n, self.E, self.G, alpha, beta, gamma)
            rotated_rho_t.append(new_rho)
        # Flip this back to the structure of rho_t
        self.rho_t = np.transpose(rotated_rho_t)[0]
                
    
    def timeEvolution(self, time, beam_profile_averaging = None, doppler_averaging = None, 
                     pretty_print_eq = None, print_eq = None, rabi_scaling = None,
                     atomic_velocity = None, r_sigma = None, n_beam_averaging = None, 
                     doppler_width = None, doppler_detunings = None):
        """ Evolves the laser-atom system over time.
        """
        n = self.n
        E = self.E
        G = self.G
        Q = self.Q
        tau = self.tau
        laser_power = self.laser_power
        laser_intensity = self.laser_intensity
        laser_wavelength = self.laser_wavelength
        rho_0 = self.rho_0
        tau_f = self.tau_f
        
        # If rho0 is not populated then set equal ground state populations
        if(not rho_0.any()):
            print("Populating ground states equally as the initial condition.")
            population = 1/len(G)
            for g in G:
                self.setRho_0(g, g, population)
        
        # Resize rho_t
        self.rho_t = [ [0 for j in range(len(time))] for i in range(self.n*self.n)]
        
        if((beam_profile_averaging) and (doppler_averaging)):
            if(laser_power):
                te.timeEvolutionGaussianAndDopplerAveraging(n, E, G, Q, self.Q_decay, tau, laser_power, r_sigma, n_beam_averaging, laser_wavelength, doppler_width, doppler_detunings, time, rho_0, self.rho_t, tau_f = tau_f, rabi_scaling = rabi_scaling, print_eq = print_eq, pretty_print_eq = pretty_print_eq, atomic_velocity = atomic_velocity)
            else:
                print("Need to have laser_power attribute in LaserAtomSystem to use beam profile avergaing! Equate <LaserAtomSystem>.laser_power to a power in milliWatts.")
        
        elif(beam_profile_averaging):
            if(laser_power):
                te.timeEvolutionGaussianAveraging(n, E, G, Q, self.Q_decay, tau, laser_power, r_sigma, n_beam_averaging, laser_wavelength, time, rho_0, self.rho_t, tau_f = tau_f, rabi_scaling = rabi_scaling, print_eq = print_eq, pretty_print_eq = pretty_print_eq, atomic_velocity = atomic_velocity)
            else:
                print("Need to have laser_power attribute in LaserAtomSystem to use beam profile avergaing! Equate <LaserAtomSystem>.laser_power to the power of the laser in mW.")
            
        elif(doppler_averaging):
            if(laser_intensity):
                te.timeEvolutionDopplerAveraging(n, E, G, Q, self.Q_decay, tau, laser_intensity, laser_wavelength, doppler_width, doppler_detunings, time, rho_0, self.rho_t, tau_f = tau_f, rabi_scaling = rabi_scaling, print_eq = print_eq, pretty_print_eq = pretty_print_eq, atomic_velocity = atomic_velocity)
            else: 
                print("Need to have laser_intensity attribute in LaserAtomSystem! Equate <LaserAtomSystem>.laser_intensity to the intensity of the laser in mW/mm^2.")

        else:
            if(laser_intensity):
                te.timeEvolution(n, E, G, Q, self.Q_decay, tau, laser_intensity, laser_wavelength, time, rho_0, self.rho_t, tau_f = tau_f, rabi_scaling = rabi_scaling, print_eq = print_eq, pretty_print_eq = pretty_print_eq, atomic_velocity = atomic_velocity)
            else: 
                print("Need to have laser_intensity attribute in LaserAtomSystem! Equate <LaserAtomSystem>.laser_intensity to the intensity of the laser in mW/mm^2.")
            
            


'''
Defines functions to symbolically print the equations of motion of the laser-atom system
Author: Manish Patel
Date created: 07/06/2021
'''

from state import *
from detuning import *
from sympy import *
from sympy import Symbol
from half_rabi_freq import *
from index import *


def symbolicPrintSystem(n, E, G, Q, Q_decay, tau_f, laser_wavelength, atomic_velocity):
    """Prints the equations of motion of the laser-atom system in full using Sympy.
    """
    symbolicPrintRhogg(n, E, G, Q, Q_decay)
    symbolicPrintRhoee(n, E, G, Q, Q_decay, tau_f)
    symbolicPrintRhoge(n, E, G, Q, Q_decay, tau_f, laser_wavelength, atomic_velocity)
    symbolicPrintRhoeg(n, E, G, Q, Q_decay, tau_f, laser_wavelength, atomic_velocity)
    
    
def symbolicPrintRhogg(n, E, G, Q, Q_decay):
        """Prints the density matrix elements rho_gg'' for the motion of the laser-atom system using Sympy.
        """
        # rho_gg''
        for g in G:  # Start with looping over g and g'' for rho_gg''
            for gpp in G:
                rho_dot = 0
                if(detuning(g, gpp) != 0):
                    rho_dot += S('Delta_{}{}*rho_{}{}'.format(g.label, gpp.label, g.label, gpp.label))
                for e in E:
                    for q in Q: 
                        if(coupling(e, gpp, q) != 0):
                            rho_dot += S('I*Omega({}, {}, {})*rho_{}{}'.format(e.label, gpp.label, q, g.label, e.label))
                for e in E:
                    column = index(e, gpp, n)
                    for q in Q: 
                        if(coupling(e, g, q) != 0):
                            rho_dot += S('-I*Omega({}, {}, {})*rho_{}{}'.format(e.label, g.label, q, e.label, gpp.label))
                for ep in E:
                    for epp in E:
                        column = index(epp, ep, n)
                        column2 = index(ep, epp, n)
                        sum_decay_channels = 0
                        for gp in G:
                            for qp in Q_decay:  # Sum over decay channel polarisations
                                sum_decay_channels += coupling(epp, gp, qp)*coupling(ep, gp, qp)
                        if(sum_decay_channels != 0):
                            for qp in Q_decay:
                                rho_dot += S('{}/(2*tau)*rho_{}{} + {}/(2*tau)*rho_{}{}'.format(coupling(ep, gpp, qp)*coupling(epp,g, qp)/sum_decay_channels,epp.label, ep.label,
                                                                             coupling(epp, gpp, qp)*coupling(ep, g, qp)/sum_decay_channels, ep.label, epp.label))
                display(Eq(S('rhodot_{}{}'.format(g.label, gpp.label)), rho_dot))

def symbolicPrintRhoee(n, E, G, Q, Q_decay, tau_f):
        """Prints the density matrix elements rho_ee'' for the motion of the laser-atom system using Sympy.
        """ 
        # rho_ee''
        for e in E:
            for epp in E:
                rho_dot = S('-1/tau*rho_{}{}'.format(e.label, epp.label))
                if(tau_f != None):
                    rho_dot += S('-rho{}{}/tau_f'.format(e.label, epp.label))
                if(detuning(e, epp) != 0):
                    rho_dot += S('Delta_{}{}*rho_{}{}'.format(e.label, epp.label, e.label, epp.label))
                for g in G:
                    for q in Q:
                        if(coupling(epp, g, q) != 0):
                            rho_dot += S('I*Omega({}, {}, {})*rho_{}{}'.format(epp.label, g.label, q, e.label, epp.label))
                for g in G:
                    for q in Q: 
                        if(coupling(e, g, q) != 0):
                            rho_dot += S('-I*Omega({}, {}, {})*rho_{}{}'.format(e.label, g.label, q, g.label, epp.label))
                display(Eq(S('rhodot_{}{}'.format(e.label, epp.label)), rho_dot))
                
def symbolicPrintRhoge(n, E, G, Q, Q_decay, tau_f, laser_wavelength, atomic_velocity):
        """Prints the density matrix elements rho_ge for the motion of the laser-atom system using Sympy.
        """ 
        # rho_ge
        for g in G:
            for e in E: 
                rho_dot = S('-rho_{}{}/(2*tau)'.format(g.label, e.label))
                if(tau_f != None):
                    rho_dot += S('-rho{}{}/(2*tau_f)'.format(g.label, e.label))
                if(dopplerDelta(e, g, w_q = 0, lambda_q = laser_wavelength, v_z = atomic_velocity) != 0):
                    rho_dot += S('-I*Delta({}, {}, omega_q, v_z)'.format(e.label, g.label))
                for ep in E:
                    for q in Q: 
                        if(coupling(ep, g, q) != 0):
                            rho_dot += S('-I*Omega({}, {}, {})*rho_{}{}'.format(ep.label, g.label, q, ep.label, e.label))
                for gp in G:
                    for q in Q: 
                        if(coupling(e, gp, q) != 0):
                            rho_dot += S('-I*Omega({}, {}, {})*rho_{}{}'.format(e.label, gp.label, q, g.label, gp.label))
                display(Eq(S('rhodot_{}{}'.format(g.label, e.label)), rho_dot))
                
def symbolicPrintRhoeg(n, E, G, Q, Q_decay, tau_f, laser_wavelength, atomic_velocity):
        """Prints the density matrix elements rho_eg for the motion of the laser-atom system using Sympy.
        """ 
        # rho_eg
        for e in E:
            for g in G:
                rho_dot = S('-rho_{}{}/(2*tau)'.format(e.label, g.label))
                if(tau_f != None):
                    rho_dot += S('-rho{}{}/(2*tau_f)'.format(e.label, g.label))
                if(dopplerDelta(e, g, w_q = 0, lambda_q = laser_wavelength, v_z = atomic_velocity) != 0):
                    rho_dot += S('-I*Delta({}, {}, omega_q, v_z)'.format(e.label, g.label))
                for ep in E:
                    for q in Q: 
                        if(coupling(ep, g, q) != 0):
                            rho_dot += S('I*Omega({}, {}, {})*rho_{}{}'.format(ep.label, g.label, q, e.label, ep.label))
                for gp in G:
                    for q in Q: 
                        if(coupling(e, g, q) != 0):
                            rho_dot += S('I*Omega({}, {}, {})*rho_{}{}'.format(e.label, g.label, q, gp.label, g.label))
                display(Eq(S('rhodot_{}{}'.format(e.label, g.label)), rho_dot))


'''
This is a file to define a function to populate the time evolution matrix for a laser-atom system
Author: Manish Patel
Date created: 12/05/2021
'''

import numpy as np
from state import *
from detuning import *
from sympy import *
from sympy import Symbol
from symbolic_print import *
from half_rabi_freq import *
from index import *

def timeEvolutionMatrix(n, E, G, Q, Q_decay, tau, laser_wavelength, laser_intensity,
                        tau_f = None, symbolic_print = None, numeric_print = None,
                       rabi_scaling = None, atomic_velocity = None):
    '''
    Function to create and populate the coupled differential equation matrix A for the laser-atom system.
    '''
    # Initialise matrix with zeros
    A = np.zeros((n*n,n*n), dtype = np.complex)
    
    # Calculate half-Rabi frequency
    rabi = halfRabiFreq(laser_intensity, tau, laser_wavelength)
    if(rabi_scaling != None):
        rabi = rabi*rabi_scaling
    
    # Initialise null parameters
    if(atomic_velocity == None):
        atomic_velocity = 0
    
    # Populate A matrix
    rho_ggpp(A, n, E, G, Q, Q_decay, tau, rabi, numeric_print = numeric_print)
    rho_eepp(A, n, E, G, Q, Q_decay, tau, rabi, tau_f = tau_f, numeric_print = numeric_print)
    rho_ge(A, n, E, G, Q, Q_decay, tau, rabi, laser_wavelength, atomic_velocity, tau_f = tau_f, numeric_print = numeric_print)
    rho_eg(A, n, E, G, Q, Q_decay, tau, rabi, laser_wavelength, atomic_velocity, tau_f = tau_f, numeric_print = numeric_print)

    # Symbolic Printing
    if(symbolic_print == True):
        init_printing()
        symbolicPrintSystem(n, E, G, Q, Q_decay, tau_f, laser_wavelength, atomic_velocity)
    
    return A

def rho_ggpp(A, n, E, G, Q, Q_decay, tau, rabi, numeric_print = None):
    # rho_gg''
    for g in G:  # Start with looping over g and g'' for rho_gg''
        for gpp in G:
            row = index(g, gpp, n)  # matrix positions of rho_gg'' in 1D array
            A[row, row] += -1.j*detuning(g, gpp)  # first term in equation
            for e in E:
                column = index(g, e, n)
                for q in Q:  # Sum over all polarisations
                    A[row, column] += coupling(e, gpp, q)*1.j*rabi
            for e in E:
                column = index(e, gpp, n)
                for q in Q: 
                    A[row, column] += -1.j*coupling(e, g, q)*rabi
            for ep in E:
                for epp in E:
                    column = index(epp, ep, n)
                    column2 = index(ep, epp, n)
                    sum_decay_channels = 0
                    for gp in G:
                        for qp in Q_decay:  # Sum over decay channel polarisations
                            sum_decay_channels += coupling(epp, gp, qp)*coupling(ep, gp, qp)
                    if(sum_decay_channels != 0):
                        for qp in Q_decay:
                            A[row, column] += 1/(2*tau)*coupling(ep, gpp, qp)*coupling(epp,g, qp)/sum_decay_channels
                            A[row, column2] += 1/(2*tau)*coupling(epp, gpp, qp)*coupling(ep, g, qp)/sum_decay_channels
            if(numeric_print == True):  # Print numerical equations
                print("rho_dot", g.label, gpp.label, " = ")
                for line in range(n*n):
                    if (A[row, line] != 0):
                        print(A[row, line], "rho", getStateLabelsFromLineNo(line, n))

def rho_eepp(A, n, E, G, Q, Q_decay, tau, rabi, tau_f = None, numeric_print = None):
    # rho_ee''
    for e in E:
        for epp in E:
            row = index(e, epp, n)
            A[row, row] += -1.j*detuning(e, epp) - 1/(tau)
            if(tau_f != None):
                A[row, row] -= 1/tau_f
            for g in G:
                column = index(e, g, n)
                for q in Q:
                    A[row, column] += 1.j*coupling(epp, g, q)*rabi
            for g in G:
                column = index(g, epp, n)
                for q in Q: 
                    A[row, column] += -1.j*coupling(e, g, q)*rabi
            if(numeric_print == True):
                print("rho_dot", e.label, epp.label, " = ")
                for line in range(n*n):
                    if (A[row, line] != 0):
                        print(A[row, line], "rho", getStateLabelsFromLineNo(line, n))

def rho_ge(A, n, E, G, Q, Q_decay, tau, rabi, laser_wavelength, atomic_velocity, tau_f = None, numeric_print = None):             # rho_ge
    for g in G:
        for e in E: 
            row = index(g, e, n)
            A[row, row] += -1.j*dopplerDelta(e, g, w_q = 0, lambda_q = laser_wavelength, 
                                                 v_z = atomic_velocity)  - 1/(2*tau) 
            if(tau_f != None):
                A[row, row] -= 1/(2*tau_f)
            for ep in E:
                column = index(ep, e, n)
                for q in Q: 
                    A[row, column] += -1.j*coupling(ep, g, q)*rabi
            for gp in G:
                column = index(g, gp, n)
                for q in Q: 
                    A[row, column] += 1.j*coupling(e, gp, q)*rabi
            if(numeric_print == True):
                print("rho_dot", g.label, e.label, " = ")
                for line in range(n*n):
                    if (A[row, line] != 0):
                        print(A[row, line], "rho", getStateLabelsFromLineNo(line, n))

def rho_eg(A, n, E, G, Q, Q_decay, tau, rabi, laser_wavelength, atomic_velocity, tau_f = None, numeric_print = None):              # rho_eg
    for e in E:
        for g in G:  
            row = index(e, g, n)
            A[row, row] += 1.j*dopplerDelta(e, g, w_q = 0, lambda_q = laser_wavelength,
                                               v_z = atomic_velocity)  - 1/(2*tau)
            if(tau_f != None):
                A[row, row] -= 1/(2*tau_f)
            for ep in E:
                column = index(e, ep, n)
                for q in Q: 
                    A[row, column] += 1.j*coupling(ep, g, q)*rabi
            for gp in G:
                column = index(gp, g, n)
                for q in Q: 
                    A[row, column] += -1.j*coupling(e, gp, q)*rabi
            if(numeric_print == True):
                print("rho_dot", e.label, g.label, " = ")
                for line in range(n*n):
                    if (A[row, line] != 0):
                        print(A[row, line], "rho", getStateLabelsFromLineNo(line, n))


'''
Class definition for an atomic state
'''

# Class to hold information to construct an atomic state
class State():
    # Parameterised constructor
    def __init__(self, L, S, m, w, label):
        self.L = L  # Orbital angular momentum quantum number
        self.S = S  # Spin quantum number
        self.m = m  # m_J degeneracy of total angular momentum
        self.J = L+S  # Total angular momentum
        self.w = w  # angular frequency of state
        self.label = label  # number labelling of state e.g. state |1> would have label 1


from setuptools import setup, find_packages

VERSION = '0.2' 
DESCRIPTION = 'LASED'
LONG_DESCRIPTION = 'A Laser-Atom Interaction Simulator using Quantum Electrodynamics'

# Setting up
setup(
       # the name must match the folder name 'LASED'
        name = "LASED", 
        version = VERSION,
        author = "Manish Patel",
        author_email="<mvpmanish@gmail.com>",
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=[
            "numpy",
            "math",
            "copy",
            "scipy",
            "sympy"
        ], # add any additional packages that 
        # needs to be installed along with your package. Eg: 'caer'
        
        keywords=['python', 'laser-atom', 'simulation', 'quantum', 'quantum electrodynamics', 'physics'],
        classifiers= [
            "Development Status :: 1 - Planning",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Microsoft :: Windows",
            "Natural Language :: English"
        ]
)


'''
Definition of the index function
'''
from state import *

def index(i, j, n):
    '''
    Function to return the index of an element of a 2D square array of 
    dimension n and indices [i,j] if converted to a 1D array
    parameters are States and the number of States in the system
    '''
    return ((i.label-1)*n+j.label)-1  # index starts at 0


def getStateLabelsFromLineNo(line_no, n):
    '''
    Gets the indeces for labelling rho [i, j] from the line number of an array
    Input: line number of an array e.g. position of row in A or column in rho_t
    Return: A tuple of [i, j]
    '''
    division = int((line_no + 1)/n)  # Gives label i
    remainder = (line_no + 1)%n  # Gives label j
    if (remainder == 0):
        i = division
        j = n
    else:
        i = (division+1)
        j = remainder
    return (i,j)